

/*
 *  functions.c - some functions used all over the place
 * This is a two-part package. Part 1 contains more
 * image-related functions and helper functions, and
 * part 2 (this one) contains more numerical algorithms and methods.
 */


/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */


#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <fftw3.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_rng.h>


#include "defs.h"
#include "functions.h"
#include "functions2.h"
#include "dlgbox.h"
#include "imgbuf.h"








/* Nifty integer square root lifted from NIH image
   (The algorithm used by this integer square root function is described
	in the document "square-root-code.txt" in the /pub/nih-image/documents
	directory on rsbweb.nih.gov.) */


/* Hmmm....do we actually still use this anywhere? */

int isqrt (long x)
{
long r, nr, m;

	r=0;
	m=0x40000000;
	do
	{
		nr = r + m;
		if (nr <= x)
		{
			x -= nr;
			r = nr + m;
		}
		r = r >> 1;
		m = m >> 2;
	}
	while (m);
	if (x > r) r++;
	return r;
}


/* Return the Gaussian function at value r - or return the second derivative of the Gaussian
	at value r. Note: the Web offers several different notions of the second derivative.
	One is (r^2-2s^2)/(2*pi*s^6) * exp (-r^2/2s^2), see http://academic.mu.edu/phys/matthysd/web226/Lab02.htm
	My own is (r^2-s^2)/N * exp (-r^2/2s^2) where s is Sigma, and N is s^5*sqrt(2*pi)
	Note also that I intentionally chaged the sign to have the tip of the Mexican hat point up.
*/

float Gauss (float r, float sigma, int LoG)
{
	if (LoG)
		//return (r*r - 2*sigma*sigma) * exp (-r*r/(2*sigma*sigma)) / (pow(sigma,4)*2.50663);
		return (2*sigma*sigma - r*r) * exp (-r*r/(2*sigma*sigma)) / (pow(sigma,5)*2.50663);
	else
		return exp (-r*r/(2*sigma*sigma)) / (sigma*2.50663);
}



/* "Pocket calculator" formula for standard deviation.
	Rounding errors allow for negative numbers in very homogeneous regions,
	so we need to trap the case where Sxx < (Sx)^2/N. IN fact, this "pocket calculator"
	method is known to be numerically unstable, so if we ever need accurate 
	variance computation, use Knuth's algorithm below. For more info, see
	http://en.wikipedia.org/wiki/Algorithms_for_calculating_variance
*/

double SDEV (double Sx, double Sxx, int N)
{
double buf;

	if (N<=1) return 0.0;			/* Too few data points */
	buf = SQR(Sx)/N;
	if (buf > Sxx) return 0.0;		/* Possibly a rounding error with SD=0 */
	buf = sqrt( (Sxx-buf)/(N-1) );
	/* dp (3, " [SDEV: Sx=%f  Sxx=%f  N=%d -> SD=%f ] ", Sx, Sxx, N, buf); */
		
	return buf;
}


double knuths_sdev (float *data, int N)		/* Compute in one pass from data */
{
double buf, mean, meansqr, delta;
int i;

	if (N<2) return 0.0;			/* Minimum of 2 data points required */
	mean = 0.0; meansqr = 0.0;

	for (i=0; i<N; i++)
	{
		buf = data[i];			/* Extract & convert to double */
		delta = buf - mean;
		mean += delta/(i+1);
		meansqr += delta*(buf-mean);
	}

	return meansqr / (N-1);
}






/* 	Peak finding algorithms. The first searches a small window near a given point
	for a local maximum.
	The second tries to find a global maximum and returns its FWHM
	data is the (float) data array (1D), and n is its length.

*/



int find_nearby_peak (float *data, int n, int srchpt)
{
int i, maxidx, wnd;
float max;

	i = srchpt; max = data[i]; maxidx = i;

	/* Search 1/10 of the data length to the left */

	wnd = n/10;
	for (i=srchpt-1; i>= MAX(0, srchpt-wnd); i--)
	{
		if (data[i] > max)
		{
			max = data[i]; maxidx=i;
		}
	}

	/* Search 1/10 of the data length to the right */

	for (i=srchpt+1; i<= MIN(n-1, srchpt+wnd); i++)
	{
		if (data[i] > max)
		{
			max = data[i]; maxidx=i;
		}
	}

	return maxidx;		/* If no maximum was found, this is the initial point */
}


float find_fwhm (float *data, int n)
{
int maxidx, i, l, r;
float max, hm, f, fwhm;

	/* Go over the data, find global maximum */

	max=data[0]; maxidx=0;
	for (i=1; i<n; i++)
	{
		if (data[i] > max)
		{
			max = data[i]; maxidx=i;
		}
	}

	/* From global maximum, go left and right, find where it drops below max/2 */

	l=maxidx-1; r=maxidx+1;
	hm = max/2.0;			/* Half maximum */
	while ( (l>=0) && (data[l] >= hm) ) l--;
	while ( (r<n)  && (data[r] >= hm) ) r++;

	/* Interpolate to find exact FWHM */

	if (l>=0) f = (hm-data[l])/(data[l+1]-data[l]); else f=0;
	fwhm = l+f;
	if (r<n) f = (hm-data[r])/(data[r-1]-data[r]); else f=0;
	fwhm = r-f - fwhm;
	return fwhm;
}





/* Interpolation coefficients for cubic interpolation. We need 4 coefficients.
	Let X be a non-integer coordinate, then the interpolation uses the four
	nearest points as follows: k is the integer part of X and x is the
	fractional part of X. The points included in the support are
	p(k-1), p(k), p(k+1), and p(k+2).

	John Russ (p.219) suggests the convolution
		p(k-1)	*	(x+3)^3 - 4(x+2)^3 + 6(x+1)^3 - 4x^3
		p(k)	*	(x+2)^3 - 4(x+1)^3 + 6x^3
		p(k+1)	*	(x+1)^3 - 4x^3
		p(k+2)	*	x^3

	The cubic spline demands the following convolution:
		p(k-1)	*	( -a x^3 + 2a x^2 - a x)
		p(k)	*	( (2-a) x^3 + (a-3) x^2 + 1 )
		p(k+1)	*	( (a-2) x^3 + (3-2a) x^2 + a x )
		p(k+2)	*	( a x^3 - a x^2 )
	where a is a parameter. Often, a=-0.5 is chosen. With a=-0.75,	
	a Hamming-windowed sinc is optimally approximated.

	Choose one of the three. Note that the Russ formulation interpolates, but
	has a worse lowpass characteristic (i.e., blurring) than linear interpolation.
	On the other hand, the Russ formulation shows no ringing, whereas the cubic
	spline does have ringing-type over- and undershoots. The reason is that Russ'
	interpolation is based on a cubic B-spline (see e.g. Dan Olfe, pg. 176) which
	means that the interpolated surface DOES NOT go through the support points.
	Conversely, the cubic spline is also known as "parabolic blend" for a=-0.5, 
	see Dan Olfe pp 163-164 for a derivation. 

	Also note that this interpolation type may be used for ANY smooth
	(that is, twice differentiable) interpolation. Example: We want to create
	smooth curve between n vertices that have an X and Y component each.
	We may use the fractional component 0<x<1 and 0<y<1 of X and Y
	to compute a set of coefficients each and use these to move smoothly
	from vertex k to vertex k+1 (additionally using vertices k-1 and k+2
	for the cubic spline)

	The parameter "type" determines the type of interpolation.
	Use 0 for the cubic B-spline, 
		1 for the cubic spline with a=0.5,
		2 for the cubic spline with a=-0.75

*/
	


void cub_coeff (double *a, double x, unsigned char type)
{
double x2, x3;
double oneover6 = 0.16666666666666667;

	switch (type)
	{
		case 0:		/* Natural B-spline, Russ p. 219 - horrible for pixel interpolation */

			a[3] = CUB(x); a[2] = CUB(1.0+x); a[1] = CUB(2.0+x);
			a[0] = (CUB(3.0+x) - 4.0*a[1] + 6.0*a[2] - 4.0*a[3]) * oneover6;
			a[1] = (6.0*a[3] - 4.0*a[2] + a[1])*oneover6;
			a[2] = (a[2] - 4.0*a[3])*oneover6;
			a[3] *= oneover6;
			break;

		case 1:						/* Cubic spline, a=-0.5 */

			x2=x*x; x3=x2*x;
			a[0] = -0.5*(x3-x2-x2+x);
			a[1] = 1.5*x3 - 2.5*x2 + 1.0;
			a[2] = -1.5*x3 + x2+x2 + 0.5*x;
			a[3] = 0.5*(x3-x2);
			break;

		case 2: 					/* Cubic spline, a=-0.75, optimum sinc-Hamming approximation */

			x2=x*x; x3=x2*x;
			a[0] = -0.75*(x3-x2-x2+x);
			a[1] = 1.25*x3 - 2.25*x2 + 1.0;
			a[2] = -1.25*x3 + 1.5*x2 + 0.75*x;
			a[3] = 0.75*(x3-x2);
			break;

		default:
			a[0] = a[1] = a[2] = a[3] = 0;	/* Try to avoid this case :-) */
	}


}


/* WINDOWING function(s) - will be extended as needed. 
	For now, we only use Hamm, Hanning, and Welch windows.
	Note that x and y need to be normalized into the range +/- 1.
	Note also that all windows except Hamming go to almost zero at the edges,
	so we create one row/column of unrecoverable data.
	If the caller normalized x and y so that it does not reach +/- 1,
	then there won't be unrecoverable data. But this is the
	caller's responsibility. */


double windowfn (double buf, double x, double y, WINDOW_TYPE wndtype)
{
double w;

	switch (wndtype)
	{
		case W_WELCH:							/* Welch */
			w = (1.0-SQR(x))*(1.0-SQR(y));
			break;
		case W_PARZEN:							/* Bartlett/Parzen */
			w = (1.0-fabs(x))*(1.0-fabs(y));
			break;
		case W_HANN:							/* Hann */
			w = 0.25*(1.0-cos(PI*(x+1.0)))*(1.0-cos(PI*(y+1.0)));
			break;
		case W_HAMMING:							/* Hamming raised cosine window */
			w = (0.54-0.46*cos(PI*(x+1.0)))*(0.54-0.46*cos(PI*(y+1.0)));
			break;
		case W_GAUSS: 							/* Gauss window with sigma=0.5 */
			w = exp(-x*x-y*y);
			break;
		default:
			w=1.0;
	}

	return buf*w;

}


/* SORTING: We need something like this all over the place.
	Implementation of the comb sort algorithm with gap adjustment
	We have two versions, one for float, one for int. 
	In both cases, we may pass a non-NULL array "links"
	which must contain integer elements, and the links
	array will be sorted synchronously with the main array
*/


int combsort_newgap (int oldgap)			/* Adjust gap size */
{
int gap; 
float shrink1 = 0.769231;		/* Reciprocal of 1.3 */
float shrink2 = 0.801711876;	/* reciprocal of 1/(1-exp(-phi)) or 1.247331 */

	gap = (int)(oldgap*shrink2);	/* Yes, we mean floor, not round */

	if( (gap == 9) || (gap == 10))
		gap = 11;					/* Gap adjustment */

	if(gap < 1) return 1;
	return gap;
}


void combsort_f (float *array, int *links, int size)
{
int gap,i,ibuf;
unsigned char swapped;
float buf;

	gap = size;

	do
	{
		swapped = 0;
		gap = combsort_newgap (gap);

		for (i=0; i< size-gap; i++)
		{
			if (array[i] > array[i+gap])
			{
				buf = array[i]; array[i] = array[i+gap]; array[i+gap] = buf;
				if (links) 		/* Swap this array, too, if non-NULL */
				{ 
					ibuf = links[i]; links[i] = links[i+gap]; links[i+gap] = ibuf; 
				}
				swapped = 1;
			}
		}

	}
	while ( (gap>1) || swapped);
}


void combsort_i (int *array, int *links, int size)
{
int gap,i,ibuf;
unsigned char swapped;
int buf, loopcnt;

	gap = size; loopcnt=0;
	dp (3, "Combsort: Array at start (size %d) is\n",size);
	for (i=0; i<size; i++) dp (3, "  %4d ",array[i]);
	dp (3, "\n");


	do
	{
		swapped = 0; loopcnt++;
		gap = combsort_newgap (gap);
		dp (3, "Combsort: New gap size is %d\n",gap);

		for (i=0; i< size-gap; i++)
		{
			if (array[i] > array[i+gap])
			{
				buf = array[i]; array[i] = array[i+gap]; array[i+gap] = buf;
				if (links) 		/* Swap this array, too, if non-NULL */
				{ 
					ibuf = links[i]; links[i] = links[i+gap]; links[i+gap] = ibuf; 
				}
				swapped = 1;
				dp (3, "Combsort: Swapping %d and %d\n",i,i+gap);
			}
		}

	}
	while ( ((gap>1) || swapped) && (loopcnt < SQR(size)*2));
	dp (3, "Combsort: Array at finish is\n");
	for (i=0; i<size; i++) dp (3, "  %4d ",array[i]);
	dp (3, "\n");

}







/* Numerical Recipes median finder. Function to help finding the n-th quantile etc.
	The input data array has n values (from 0 to n-1). This function
	returns the k-th smallest value in the float array data. The input array will be rearranged
	to have this value in location data[k], with all smaller elements moved to data[0..k-2] 
	(in arbitrary order) and all larger elements in data[k..n-1] (also in arbitrary order).
	NOTE: FUNCTION CHANGES ELEMENT ORDER OF INPUT ARRAY.
*/

#define WVSWAP(A,B) temp=(A);(A)=(B);(B)=temp;

float element_finder (float *data, int k, int n)
{
int i,rght,j,lft,mid;
float a,temp;
unsigned char abort;
	
	lft=0;
	rght=n-1;
	abort=FALSE;
	
	while (TRUE)					/* Exit this loop through the RETURN below */
	{
		if (rght <= lft+1) 				/* Active partition contains 1 or 2 elements. */
		{
			if ((rght == lft+1) && (data[rght] < data[lft])) 		/* Case of 2 elements. */
			{
				WVSWAP (data[lft], data[rght])
			}
			return data[k];
		} 
		else 
		{
			mid=(lft+rght) >> 1;					/* Choose median of left, center, and right elements... */
			WVSWAP(data[mid],data[lft+1]);			/* ... as partitioning element a. */
			
			if (data[lft] > data[rght]) 			/* Also, rearrange so that data[l]  data[l+1], data[ir]  data[l+1] */
				WVSWAP(data[lft],data[rght]);
			
			if (data[lft+1] > data[rght])
				WVSWAP(data[lft],data[rght]);
			
			if (data[lft] > data[lft+1]) 
				WVSWAP(data[lft],data[lft+1]);

			i=lft+1;								/* Initialize pointers for partitioning. */
			j=rght;
			a=data[lft+1];							/* Partitioning element. */
			abort=FALSE;
			while (!abort)							/* Begin innermost loop */
			{
				do 
					i++; 							/* Scan up to find element > a.*/
				while (data[i] < a);
				do 
					j--; 							/* Scan down to find element < a. */
				while (data[j] > a);
				if (j < i) 
					abort=TRUE;					/* Pointers crossed. Partitioning complete. */
				else
					WVSWAP(data[i],data[j]);
			}

			data[lft+1]=data[j];					/* Insert partitioning element. */
			data[j]=a;
			if (j >= k) rght=j-1;					/* Keep active the partition that contains the kth element. */
			if (j <= k) lft=i;
		}
	}
}


#undef WVSWAP




/*-------------------------------------------------------------------------*/

/* Perform a log-log fit into xdata/ydata from st to en.
 	Report the results in 
		H	= slope of fit
		intcpt = intercept
		r	= Pearson correlation

	The loglogfit function uses base-10 log.

	Can we simply log-transform the data? This is an ongoing debate. See, for example,
	Xiao X, White EP, Hooten MB, Durham SL. On the use of logtransformation vs nonlinear
	regression for analyzing biological power laws. Ecology 2011; 92: 1887--94.
*/


fitresult loglogfit (double* x, double* y, int st, int en)
{
double lgx,lgy,r,Sx,Sxx,Sy,Syy,Sxy,H,intcpt;
int i,j;
fitresult result;

	Sx=Sxx=Sy=Syy=Sxy=0; j=0;
	for (i=st; i<=en; i++)
	{
		lgx = SAFELOG10(x[i]); lgy = SAFELOG10(y[i]);
		Sx  += lgx;
		Sxx += SQR(lgx);
		Sy  += lgy;
		Syy += SQR(lgy);
		Sxy += lgx*lgy;
		j++;
	}
	H = (Sx*Sy - j*Sxy)/(Sx*Sx - j*Sxx);
	intcpt = (Sy-H*Sx)/j;
	r = (j*Sxy-Sx*Sy)/(sqrt(j*Sxx-Sx*Sx)*sqrt(j*Syy-Sy*Sy));

	result.H = H;
	result.intcpt=intcpt;
	result.r=r;
	return result;

}


/* Perform a linear LSQ fit into xdata/ydata from st to en.
 	Report the results in H, intcpt and r, 
*/

fitresult lsqfit (double* xdata, double* ydata, int st, int en)
{
double r,Sx,Sxx,Sy,Syy,Sxy,H,intcpt;
int i,j;
fitresult result;

	Sx=Sxx=Sy=Syy=Sxy=0; j=0;
	for (i=st; i<=en; i++)
	{
		Sx  += xdata[i];
		Sxx += SQR(xdata[i]);
		Sy  += ydata[i];
		Syy += SQR(ydata[i]);
		Sxy += xdata[i]*ydata[i];
		j++;
	}
	H = (Sx*Sy - j*Sxy)/(Sx*Sx - j*Sxx);
	intcpt = (Sy-H*Sx)/j;
	r = (j*Sxy-Sx*Sy)/(sqrt(j*Sxx-Sx*Sx)*sqrt(j*Syy-Sy*Sy));

	result.H = H;
	result.intcpt=intcpt;
	result.r=r;
	return result;

}



/******************************************************

	NATURAL CUBIC SPLINES

******************************************************/

/* If there was ever a book as close to perfect when it comes
	to computer graphics, it is Dan Olfe's Computer Graphics for Design
	(Prentice Hall, 1995). Of course, it is out of print :-(
	The natural cubic spline is explained in this book on pp. 166-169.

	In this section, we implement curve interpolation through natural
	cubic splines. Curves can be 1D, 2D or 3D, and each dimension needs
	its own spline_block. For example, we have a curve with N vertices
	x[0],y[0] through x[N-1],y[N-1]. Each segment of the curve, that is,
	x[i],y[i] to x[i+1],y[i+1] is connected by a cubic function,
	
				x(i,t) = ax[i] + bx[i]*t + cx[i]*t^2 + dx[i]*t^3
				y(i,t) = ay[i] + by[i]*t + cy[i]*t^2 + dy[i]*t^3

	where x(i,t),y(i,t) is the interpolated point along the segment,
	t is the position of the point (t goes from 0 to 1 as the curve moves
	from vertex i to vertex i+1), and ax...dx are the coefficient vectors
	for x. Similarly, ay...dy are the coefficient vectors for y, created 
	independently with spline_new.

	For each dimension, we need to first call spline_new and pass the 
	vertices v[i] and the number of points. spline_new then creates
	the coefficient vectors and returns them in spl.

	To get an interpolated point, we pass the spline_block to spline_get,
	and we specify the segment (i) and the position along the segment t.
	The function returns one interpolated value. For a 2D vertex, we need
	to call the function twice: once for x and once for y.

	Since spline_new uses malloc, we need to use spline_free to free the
	memory allocated with spline_new.
*/



/* Dan Olfe presents a clever recursive computation scheme for the
	spline coeficcients a,b,c,d, which we use here. */

void spline_new (spline_block *spl, int numpts, float *v)
{
int i,m;
float *gamma;
float *delta;
float *bigd;

	m = numpts-1;
	spl->npoints = numpts-1;
	spl->coeff_a = calloc (numpts+2, sizeof(float));
	spl->coeff_b = calloc (numpts+2, sizeof(float));
	spl->coeff_c = calloc (numpts+2, sizeof(float));
	spl->coeff_d = calloc (numpts+2, sizeof(float));
	gamma = calloc (numpts+2, sizeof(float));
	delta = calloc (numpts+2, sizeof(float));
	bigd = calloc (numpts+2, sizeof(float));

	/* Calculate the gamma coefficients */

	gamma[0] = 0.5;									/* left endpoint */
	for (i=1; i<m; i++)
		gamma[i] = 1.0/(4.0 - gamma[i-1]);

	gamma[m] = 1.0/(2.0 - gamma[m-1]);				/* right endpoint */


	/* Recursively calculate the delta vector; builds on gamma. */

	delta[0] = 3.0*gamma[0]*(v[1]-v[0]);
	for (i=1; i<m; i++)
		delta[i] = gamma[i]*( 3.0*(v[i+1]-v[i-1]) - delta[i-1]);

	delta[m] = gamma[m]*( 3.0*(v[m]-v[m-1]) - delta[m-1]);

	/* Recusively calculate the diagonal elements of the D matrix,
		these are already the coefficients b. */

	bigd[m] = delta[m];
	for (i=m-1; i>=0; i--)
		bigd[i] = delta[i] - gamma[i]*bigd[i+1];

	/* Last step: Create the coefficients a,b,c,d from D */

	dp (3, "------------------------------------------------------------------\n");
	dp (3, "Vtx   vertex    gamma     delta     big D     coeff c   coeff_d\n");
	for (i=0; i<m; i++)			/* We don't carry the last vertex, see also the spl->npoints assigment */
	{
		spl->coeff_a[i] = v[i];			/* Continuity: coincide with vertex when t=0 */
		spl->coeff_b[i] = bigd[i];			/* Extra step, but increases readability */
		spl->coeff_c[i] = 3.0*(v[i+1]-v[i]) - 2.0*bigd[i] - bigd[i+1];
		spl->coeff_d[i] = 2.0*(v[i]-v[i+1]) + bigd[i] + bigd[i+1];

		dp (3, "%3d   %6.1f    %6.1f    %6.1f    %6.1f    %6.1f    %6.1f\n", i,v[i], gamma[i], delta[i], bigd[i], spl->coeff_c[i], spl->coeff_d[i]);

	}
	dp (3, "------------------------------------------------------------------\n");

	/* Free the intermediate arrays */

	free (gamma); free (delta); free (bigd);

}


float spline_get (spline_block *spl, int segment, float t)
{
float a,b,c,d, r;

	if (segment > spl->npoints)
	{
		dp (0, "Cubic spline: Attempt to access invalid segment %d\n", segment);
		return 0.0;
	}
	if ((t<0.0) || (t>1.0))
	{
		dp (0, "Cubic spline: Attempt to access invalid fraction %f of segment %d\n", t, segment);
		return 0.0;
	}

	a = spl->coeff_a[segment];			/* This is to improve readability */
	b = spl->coeff_b[segment];
	c = spl->coeff_c[segment];
	d = spl->coeff_d[segment];

	r = a + t*(b + t*(c + t*d));		/* Interpolation function */
	return r;

}


void spline_free (spline_block *spl)
{

	spl->npoints = 0;
	if (spl->coeff_a) free (spl->coeff_a);
	spl->coeff_a = NULL;
	if (spl->coeff_b) free (spl->coeff_b);
	spl->coeff_b = NULL;
	if (spl->coeff_c) free (spl->coeff_c);
	spl->coeff_c = NULL;
	if (spl->coeff_d) free (spl->coeff_d);
	spl->coeff_d = NULL;

}



/******************************************************

	MATRIX FUNCTIONS and related stuff

******************************************************/


/* det2x2 returns the determinant of a 2x2 matrix, quick n simple.
	A is supposed to have four elements:
	A[0] is a11, A[1] is a12, A[2] is a21, A[3] is a22. */

MFLOAT det2x2 (MFLOAT *A)
{
	return (A[0]*A[3] - A[1]*A[2]);
}



/* Pythag (numerical Recipes) 
   Computes (a**2 + b**2 )**1/2 without destructive under- or overflow.
*/


MFLOAT pythag (MFLOAT a, MFLOAT b) 
{ 
MFLOAT absa, absb; 

	absa=fabs(a); absb=fabs(b); 
	if (absa > absb) 
		return absa*sqrt(1.0+SQR(absb/absa)); 
	else 
		return (absb == 0.0 ? 0.0 : absb*sqrt(1.0+SQR(absa/absb)));
}


void fswap (MFLOAT* a, MFLOAT* b)
{
MFLOAT c;

	c=*a; *a=*b; *b=c;
}


/* Multiply a n by n matrix A with a n-vector V. Upon exit, the
   result is in V */

void matmult (MFLOAT* A, MFLOAT* V, int n)
{
int i,j;
MFLOAT R[n];

	for (i=0; i<n; i++)
	{
		R[i]=0;
		for (j=0; j<n; j++) R[i]+=V[j]*A[j+n*i];
	}
	for (i=0; i<n; i++) V[i]=R[i];
}



/* Multiply a n by m matrix A with a n-vector V. Upon exit, the
   result is in R, a m-vector, allocated by the caller */

void matmulnm (MFLOAT *R, MFLOAT* A, MFLOAT* V, int n, int m)
{
int i,j;

	for (i=0; i<m; i++)
	{
		R[i]=0;
		for (j=0; j<n; j++) R[i]+=V[j]*A[j+n*i];
	}
}



/* Solve a linear equation system by Gauss-Jordan Elimination.
   This function accepts a matrix A (n by n elements, defined by
   index col + n*row) and a vector B (n elements). It returns
   the solution vector X in B and the inverse of A in A */


int matsolv (MFLOAT *A, MFLOAT *B, int n) 
{ 
int *colidx,*rowidx,*ipiv; 
int i,col,row,j,k,l,ll, pividx; 
MFLOAT max, buf, pivinv, temp; 

	/* Please note that we swap rows/cols in this function, effectively
	   working on the transposed matrix. The caller needs to prepare
	   the matrix the usual (for me usual) way, i.e. x-index is the
	   fast index, while this algorithm assumes the Y index as the fast
	   index. This is caused by the original way of handling a matrix
	   using an array of pointers, i.e. A[n][n]. */


	colidx = calloc (n, sizeof (int));
	rowidx = calloc (n, sizeof (int));
	ipiv = calloc (n, sizeof (int)); /* Sic, calloc. We need to init this with 0 */
	
	row=col=0; /* silence a compiler warning */
	
	for (i=0; i<n; i++) 
	{ 
		
		/* Look for a pivot element. It is the matrix' maximum that has not yet been pivoted */
		
		max=0.0;
		for (j=0; j<n; j++) 
		{
			if (ipiv[j] != 1) 
				for (k=0; k<n; k++) 
				{ 
					if (ipiv[k] == 0) 
					{ 
						if (fabs(A[n*j+k]) >= max) 
						{ 
							max = fabs(A[n*j+k]); 
							row=j; col=k; 
						} 
					} 
				} 
		}
		(ipiv[col])++; 
		
		/* row / col now indicate the location of the pivot element */
		
		if (row != col) 
		{ 
			for (l=0; l<n; l++) { temp=A[row*n+l]; A[row*n+l]=A[col*n+l]; A[col*n+l]=temp; }
			temp=B[row]; B[row]=B[col]; B[col]=temp;
		} 
		
		/* Divide the pivot row by the pivot element */
		
		rowidx[i]=row; 
		colidx[i]=col; 
		pividx = col + n*col;
		if (A[pividx] == 0.0) return -1; /* Singular matrix */
		pivinv=1.0/A[pividx]; A[pividx]=1.0; 
		for (l=0; l<n; l++) A[col*n+l] *= pivinv; 
		B[col] *= pivinv;
		for (ll=0; ll<n; ll++) 
			if (ll != col) 
			{ 
				buf=A[ll*n+col]; 
				A[ll*n+col]=0.0; 
				for (l=0; l<n; l++) A[ll*n+l] -= A[col*n+l]*buf; 
				B[ll] -= B[col]*buf; 
			} 
	}
	
	for (l=n-1; l>=0; l--) 
	{ 
		if (rowidx[l] != colidx[l]) 
			for (k=0; k<n; k++) 
			{
				temp = A[k*n+rowidx[l] ]; 
				A[k*n+rowidx[l] ] = A[k*n+colidx[l] ];
				A[k*n+colidx[l] ] = temp;
			}
	} 
				
	free (ipiv); 
	free (rowidx); 
	free (colidx); 
	return 0;

}





void mat_transpose (MFLOAT *A, int n)
{
int x,y;
MFLOAT temp;

	for (y=0; y<n; y++)
		for (x=y+1; x<n; x++)
			if (x!=y)
			{
				temp = A[ x+n*y ];
				A[ x+n*y ] = A[ y+n*x ];
				A[ y+n*x ] = temp;
			}

}


/***************************************************************

Householder reduction of a real, symmetric matrix A.
A is passed as linear n by n array, and we use the macro MA
to access the elements.
On output, A is replaced by the orthogonal matrix Q effecting the
transformation. 
D[0..n-1] returns the diagonal elements of the tridiagonal matrix,
and E[0..n-1] the the off-diagonal elements with e[0]=0;
*/

#define MA(X,Y)		A[(X)+n*(Y)]
/*#undef MFLOAT
#define MFLOAT float*/

void householder_reduction (MFLOAT* A, int n, MFLOAT* D, MFLOAT* E)
{ 
int i,j,k,l;
MFLOAT scale,hh,h,g,f; 


	for (i=n-1; i>=1; i--) 
	{ 
		l=i-1; 
		h=0.0; scale=0.0; 
		if (l > 0)
		{ 
			for (k=0; k<=l; k++) 
				scale += fabs(MA(i,k)); 

			if (scale == 0.0) 	/* Skip transformation. */
				E[i]=MA(i,l); 
			else 
			{ 
				for (k=0; k<=l; k++) 
				{ 
					MA(i,k) /= scale; 		/* Use scaled A's for transformation */
					h += MA(i,k)*MA(i,k); 	/* Form sigma in h */
				} 
				
				f=MA(i,l);
				g = (f >= 0.0 ? -sqrt(h) : sqrt(h)); 
				E[i]=scale*g; 
				h -= f*g; 
				MA(i,l)=f-g; 				/* Store u in the i-th row of a  */
				
				f=0.0; 
				for (j=0; j<=l; j++) 
				{ 
					MA(j,i)=MA(i,j)/h; 		/* Store u/H in i-th column of a  */

					g=0.0; 					/* Form a element of A � u in g */
					for (k=0; k<=j; k++)  g += MA(j,k)*MA(i,k);
					for (k=j+1; k<=l; k++)  g += MA(k,j)*MA(i,k);
					E[j]=g/h; 				/* Form element of p in temporarily unused element of e */
					f += E[j]*MA(i,j);
				}
				
				hh=f/(h+h);
				for (j=0; j<=l; j++) 
				{ 
					f=MA(i,j);
					g=E[j]-hh*f; 
					E[j] = g;
					for (k=0; k<=j; k++) MA(j,k) -= (f*E[k]+g*MA(i,k)); 
				} 
			} 
		}
		else 
			E[i]=MA(i,l);
		
		D[i]=h; 
	}
	
	D[0]=0.0; E[0]=0.0; 
	for (i=0; i<n; i++) 
	{ 				/* Begin accumulation of transformation matrices. */
		l=i-1; 
		if (D[i]) 
		{ 							/* This block skipped when i==0 */
			for (j=0; j<=l; j++)
			{
				g=0.0; 
				for (k=0; k<=l; k++) 	/* Use u and u/H stored in A to form P � Q */
					g += MA(i,k)*MA(k,j);
				for (k=0; k<=l; k++) 
					MA(k,j) -= g*MA(k,i);
			} 
		} 

		D[i]=MA(i,i);
		MA(i,i)=1.0; 	/* Reset row and colum of A to identity matrix for next iteration. */

		for (j=0; j<=l; j++) MA(j,i)=MA(i,j)=0.0; 

	}

}


/*
QL algorithm with implicit shifts, to determine the eigenvalues 
and eigenvectors of a real, symmetric, tridiagonal matrix, 
or of a real, symmetric matrix previously reduced by Householder reduction.
On input, D[0..n-1] contains the diagonal elements of the 
tridiagonal matrix. On output, it returns the eigenvalues. 
The vector E[0..n-1] inputs the subdiagonal elements of the 
tridiagonal matrix, with E[0] arbitrary. 
On output E is destroyed. 
If the eigenvectors of a tridiagonal matrix are desired, 
the matrix A (linear 0..n-1, 0..n-1) is input as the identity matrix. 
If the eigenvectors of a matrix that has been reduced by Householder
are required, then A is input as the matrix A output by Householder. 
In either case, the k-th column of A returns the 
normalized eigenvector corresponding to D[k].
NOTE THAT UPON OUTPUT A IS TRANSPOSED and sorted by D.
This is what we need for the K-L transform.
*/

#define XFRSIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))


void eigen_ql (MFLOAT* A, int n, MFLOAT* D, MFLOAT* E)
{
int m,l,iter,i,k; 
MFLOAT s,r,p,g,f,dd,c,b; 
MFLOAT max; int j, maxidx;

	for (i=1; i<n; i++) 
		E[i-1]=E[i]; 							/* Convenient to renumber the el-ements of e. */
	
	E[n-1]=0.0; 
	for (l=0; l<n; l++) 
	{ 
		iter=0; 
		do 
		{ 
			for (m=l; m<n-1; m++) 
			{ 									/* Look for a single small subdiagonal element to split the matrix. */
				dd=fabs(D[m])+fabs(D[m+1]); 
				if ((float)(fabs(E[m])+dd) == (float)dd)
				{
					dp (3, "Found small element at %d\n",m);
					break; 
				}
			} 
			if (m != l)
			{ 
				if (iter++ == 30) 
					dp (0,"Too many iterations in eigenvector iteration"); 

				g=(D[l+1]-D[l])/(2.0*E[l]); 		/* Form shift. */
				r=pythag(g,1.0); 
				g=D[m]-D[l]+E[l]/(g+XFRSIGN(r,g)); 	/* This is dm - ks. */
				s=c=1.0; p=0.0;

				for (i=m-1;i>=l;i--) 
				{ 								/* A plane rotation as in the original QL, followed by Givens rotations to restore tridiagonal form. */
					f=s*E[i]; 
					b=c*E[i]; 
					E[i+1]=(r=pythag(f,g)); 
					if (r == 0.0) 
					{ 							/* Recover from underflow. */
						D[i+1] -= p; 
						E[m]=0.0;
						dp (3, "UNDERFLOW!\n");
						break; /* break exits the for-loop */
					} 
					
					s=f/r; 
					c=g/r; 
					g=D[i+1]-p; 
					r=(D[i]-g)*s+2.0*c*b; 
					D[i+1]=g+(p=s*r); 
					g=c*r-b;
					for (k=0; k<n; k++) 	
					{ 							/* Form eigenvectors. */
						dp (3,"Before: f=A(%d,%d)=%f, A(%d,%d)=%f,\n",k,i+1,MA(k,i+1),k,i,MA(k,i));
						f=MA(k,i+1); 
						MA(k,i+1)=s*MA(k,i)+c*f; 
						MA(k,i)=c*MA(k,i)-s*f; 
						dp (3,"After : s=%f, c=%f, A(%d,%d)=%f, A(%d,%d)=%f,\n",s,c,k,i+1,MA(k,i+1),k,i,MA(k,i));
					} 
				} 
				dp (3,"***    r=%f\n",r);
				/*
				if (r == 0.0 && i >= l) continue; /* cont skips the rest and cycles the for-loop */
				if ((r!=0.0) || (i<l))
				{
					D[l] -= p; E[l]=g; E[m]=0.0; 
				}
			} 
		} 
		while (m != l); 
	}

	/* Finally, sort the rows of the transposed A by descending D */

	for (i=0; i<n-1; i++)
	{
		max=D[i]; maxidx=i;
		for (j=i+1; j<n; j++)
		{
			if (D[j] > max) { max=D[j]; maxidx=j; }
		}
		if (maxidx!=i)
		{
			r=D[i]; D[i]=D[maxidx]; D[maxidx]=r;
			for (j=0; j<n; j++)
			{
				r=MA(j,i); MA(j,i)=MA(j,maxidx); MA(j,maxidx)=r;
			}
		}
	}
}




				


/* To verify this in Octave use

	a=zeros(3,3);
	a(1,:)=[3,2,7];
	a(2,:)=[2,5,6];
	a(3,:)=[7,6,1];
	q=orth(a);
	[v,d]=eig(a);
*/


void testmatrix ()
{
int n=3;
/*MFLOAT A[9] = { 3,2,7,2,5,6,7,6,1};*/
/*MFLOAT A[9] = { 2,0,1,0,2,0,1,0,2};*/
MFLOAT A[9] = { 0.996094, 0.953125, 0.726563, 0.953125, 1.9375, 1.28125, 0.726563, 1.28125, 1.359375};
MFLOAT D[n], E[n];
int i,j;


	dp (0, "Symmetrical Matrix before:\n");

	for (j=0; j<n; j++)
	{
		for (i=0; i<n; i++)
		{
			dp (0, "%6.2f   ",MA(i,j));
		}
		dp (0,"\n");
	}

	householder_reduction (A,n,D,E);

	dp (0, "Orthogonal Matrix after reduction:\n");

	/* Should be 	0.5388    0.5266    0.6576
    				0.5859    0.3266   -0.7417
    				0.6053   -0.7849    0.1325 (matlab)
					
					0.65    -0.76     0.00
 					-0.76   -0.65     0.00
  					0.00     0.00     1.00	(n.r.)

	*/

	for (j=0; j<n; j++)
	{
		for (i=0; i<n; i++)
		{
			dp (0, "%6.2f   ",MA(i,j));
		}
		dp (0,"\n");
	}
	
	dp (0, "Diagonal elements D:     ");
	for (j=0; j<n; j++) dp (0, "%6.2f  ",D[j]);
	dp (0,"\nOff-diagonal elements E: ");
	for (j=0; j<n; j++) dp (0, "%6.2f  ",E[j]);
	dp (0,"\n");

	/* Eigenvalues are     2.1548, -6.193, 13.0384 */


	eigen_ql (A,n,D,E);
	
	dp (0, "Eigen Matrix after QL algorithm:\n");

	/* Should be 	   	0.6576   -0.5266    0.5388
   						-0.7417   -0.3266    0.5859
    					0.1325    0.7849    0.6053


	*/

	for (j=0; j<n; j++)
	{
		for (i=0; i<n; i++)
		{
			dp (0, "%6.2f   ",MA(i,j));
		}
		dp (0,"\n");
	}
	
	dp (0, "Eigenvalues D:     ");
	for (j=0; j<n; j++) dp (0, "%6.2f  ",D[j]);
	dp (0,"\n");

	/* Eigenvalues are     2.1548, -6.193, 13.0384 */

	


}



#undef MA


/***************************************************************






/******************************************************************/




