
/*
 *	Hough: All things Hough transform
 *
 */



/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.

		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */


#include <math.h>
#include <stdlib.h>
#include <stdio.h>


#include "hough.h"
#include "roi.h"
#include "dlgbox.h"
#include "imgbuf.h"
#include "application.h"
#include "secondwnd.h"
#include "filter.h"
#include "functions.h"
#include "functions2.h"
#include "randgen.h"
#include "gtkhelpers.h"



/* Standard line Hough transform in the version of Duda and Hart.
	Transform takes place slice-by-slice for all slices.
	The source image is expected to be thresholded, and contains only edges.
	Each nonzero point casts votes for angles from 0 to 180 degrees
	(180 - 360 degrees omitted for symmetry reasons)
	The transform image is finally returned through img.

	Interesting side note - the line transforms are very, very close relatives
	of the Radon transform in the CT module.
*/


/* TODO: Optionally use the edge strength (i.e., pixel value) as weight for the votes */



void hough_line (image_type *img, float deltatheta)
{
image_type hough; float *p;
float theta, rho, buf;
int x,y,z, tx,ty, rx, ry, xm, ym, tm, hslcidx;

	/* Allocate target image. Hough transform is always FLOAT */

	tx = (int)(sqrt(2)*MAX(img->xmax, img->ymax));		/* Maximum value for rho axis */
	ty = 180.0/deltatheta;				/* Maximum value for theta axis */
	tm = tx/2; xm = img->xmax/2; ym=img->ymax/2;
	hslcidx = 0;						/* Index into Hough data slice */

	if (allocate_image (&hough, B_FLOAT, tx, ty, img->zmax)==-1)
		fatal_memerr ();
	p = hough.data;

	for (z=0; z<img->zmax; z++)			/* Run over all slices */
	{
		for (y=0; y<img->ymax; y++)
			for (x=0; x<img->xmax; x++)
			{
				if (nonzero(img,x,y,z))		/* For all nonzero pixels of this slice */
				{
					ry=0;						/* angle axis index */
					buf = readbuf_flt (img,x,y,z) / ty;		/* TODO: What for? Edge weight ?? */
					for (theta=0; theta < 180; theta+=deltatheta)	/* each point, all angles */
					{
						rho = (x-xm)*COS(theta) + (y-ym)*SIN(theta);
						rx = IROUND(rho+tm);
						if ((rx>=0) && (rx < tx))
						{
							p[rx + tx*ry + hslcidx] +=1.0;			/* can use img buf value for this - intensity-weighted */
						}
						else
							dp (1, "Hough: Coordinate error at (%d,%d,%d), theta %5.1f, rho=%f, index %d\n",x,y,z,theta,rho,rx);

						ry++;			/* Next point along theta axis */
					}
				}
			}
		hslcidx += tx*ty;			/* next Hough data slice */
	}

	make_minmax (&hough);
	handback (img, &hough);

}


/* This function is a variation of the above. We combine the Hough transform with
   the Kirsch edge detector. As a consequence, we get the edge direction and can
   therefore restrict the range of theta. We get fewer, but more meaningful votes
   from each pixel. NOTE that this function operates on a single slice only.
   Important: 	1) The threshold should be known beforehand.
				2) Due to the lower number of votes, delta theta should be small
*/


void hough_line_edge (image_type *img, float deltatheta, float thresh, int slice)
{
image_type hough; float *p;
float theta, thetastart, thetaend, rho, buf;
int i,j,x,y,tx,ty, rx, ry, xm, ym, tm, m;
int ornt;
float max;
float b[4];

float kirschfac_a[3][3]= 	{	{-1.0, -2.0, -1.0} , {0.0, 0.0, 0.0}, {1.0, 2.0, 1.0}	};	/* Horizontal & vertical edge */
float kirschfac_b[3][3]=	{	{ 0.0,-1.0,-2.0} , { 1.0, 0.0,-1.0}, {+2.0, 1.0, 0.0}	};	/* 45 degrees */
float kirschfac_c[3][3]=	{	{ -2.0,-1.0,0.0} , { -1.0, 0.0,1.0}, {0.0, 1.0, 2.0}	};	/* -45 degrees */



	/* Allocate target image. Hough transform is always FLOAT */

	tx = (int)(sqrt(2)*MAX(img->xmax, img->ymax));		/* Maximum value for rho axis */
	ty = 180.0/deltatheta;				/* Maximum value for theta axis */
	tm = tx/2; xm = img->xmax/2; ym=img->ymax/2;

	if (allocate_image (&hough, B_FLOAT, tx, ty, 1)==-1)
		fatal_memerr ();
	p = hough.data;


	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
		{
			/* For this data point, determine the 4 Kirsch magnitudes */

			for (m=0; m<4; m++) b[m]=0;
			for (j=0; j<=2; j++)
			{
				for (i=0; i<=2; i++)
				{
					buf = readbuf_flt(img,x+i-1,y+j-1,slice);

					b[0] += buf * kirschfac_a[i][j];		/* Vertical edge */
					b[1] += buf * kirschfac_b[i][j];		/* 45° downward slope, 45° CW from x axis */
					b[2] += buf * kirschfac_a[j][i];		/* Horizontal edge */
					b[3] += buf * kirschfac_c[i][j];		/* 45° upward slope, 45° CCW from x axis */
				}
			}

			/* Compute final magnitude (for thresholding) and orientation */

			ornt=0; max=0;
			for (m=0; m<4; m++)
			{
				b[m] = fabs (b[m]);			/* The fabs takes care of the other 4 directions */
				if (max < b[m]) { max=b[m]; ornt=m; }
			}

			if (ornt==0)
				{ thetastart = -22; thetaend=+22; }
			else if (ornt==1)
				{ thetastart = 135-22; thetaend=135+22; }
			else if (ornt==2)
				{ thetastart = 90-22; thetaend= 90+22; }
			else /* if (ornt==3) */
				{ thetastart = 45-22; thetaend=45+22; }


			/* Time to cast the votes */

			if (max >= thresh)		/* For all pixels of this slice that lie above the threshold */
			{
					for (theta=thetastart; theta <= thetaend; theta+=deltatheta)	/* each point, restricted angles */
					{
						rho = (x-xm)*COS(theta) + (y-ym)*SIN(theta);
						rx = IROUND(rho+tm);
						ry = IROUND(theta/deltatheta); if (ry<0) ry+=ty; else if (ry>=ty) ry-=ty;
						if ((rx>=0) && (rx < tx))
						{
							p[rx + tx*ry ] +=1.0;			/* can use img buf value (i.e., max) for this - intensity-weighted */
						}
						else
							dp (1, "Hough: Coordinate error at (%d,%d), theta %5.1f, rho=%f, index %d\n",x,y,theta,rho,rx);

					}

			}
	}

	make_minmax (&hough);
	handback (img, &hough);

}




/* Randomized line Hough transform.
	Transform takes place slice-by-slice for a single slice.
	The source image is expected to be thresholded, and contains only edges.
	We randomly sample pairs of nonzero points, compute rho and theta,
	and accumulate the target image at rho,theta.
	(180 - 360 degrees omitted for symmetry reasons)
	The transform image is finally returned through img.

*/

void randomized_hough_line (image_type *img, float deltatheta, int slice)
{
image_type hough; float *p;
float theta, rho, buf;
int tx,ty, rx, ry, xm, ym, tm, maxaccu, itercnt, findcnt;
int x1,y1,x2,y2;
unsigned char zeroelem[4] = {0,0,0,0};


	/* Allocate target image. Hough transform is always FLOAT */

	tx = (int)(sqrt(2)*MAX(img->xmax, img->ymax));		/* Maximum value for rho axis */
	ty = 180.0/deltatheta;				/* Maximum value for theta axis */
	tm = tx/2; xm = img->xmax/2; ym=img->ymax/2;

	if (allocate_image (&hough, B_FLOAT, tx, ty, 1)==-1)
		fatal_memerr ();
	p = hough.data;
	dp (2, "Hough: Allocated Hough space with %d x %d accumulators\n",tx,ty);

	maxaccu = 0; itercnt=0; findcnt=0;
	while ((maxaccu < 100) && (itercnt < 10000) && (findcnt < 20000))
	{

		findcnt=0;
		do					/* Pick a nonzero point x1,y1 */
		{
			x1 = (int)(nr_rand2() * img->xmax);
			y1 = (int)(nr_rand2() * img->ymax);
			findcnt++;
		}
		while ((!nonzero(img,x1,y1,slice)) && (findcnt < 20000));

		do					/* Pick a nonzero point x2,y2 != x1,y1 */
		{
			do
			{
				x2 = (int)(nr_rand2() * img->xmax);
				y2 = (int)(nr_rand2() * img->ymax);
			}
			while ( (SQR(x1-x2)+SQR(y2-y1) < 16));		/* Make sure these are more than 4 pixels apart or angle might be inaccurate */
			findcnt++;
		}
		while ((!nonzero(img,x2,y2,slice))  && (findcnt < 20000));

		/* TODO: Shouldn't we suppress the following if selection of v1 and v2 failed??? */

		theta = (atan2 (y2-y1,x1-x2));
		rho = 0.5*((x1+x2-2*xm)*cos(theta)+(y1+y2-2*ym)*sin(theta));
		theta = (theta*180.0/PI);

		rx = IROUND(rho+tm);
		ry = (int) (theta/deltatheta); if (ry<0) ry+=ty; else if (ry>=ty) ry-=ty;
		if ((rx>=0) && (rx < tx) && (ry>=0) && (ry < ty))
		{
			buf = p[rx + tx*ry ] = p[rx + tx*ry ] + 1.0;
			if (maxaccu < buf) maxaccu = buf;					/* Update highest accumulator value */

			/* If we accepted to accumulate, delete the point pair from the original */

			writebuf (img, zeroelem, x1,y1,slice);
			writebuf (img, zeroelem, x2,y2,slice);

		}
		else
			dp (1, "Hough: Coordinate error at (%d,%d:%d,%d), theta %5.1f, rho=%f, index (%d,%d)\n",x1,y1,x2,y2,theta,rho,rx,ry);

		itercnt++;		/* Update number of iterations */
		if ( (itercnt & 0x7f) == 0)
			dp (2, "Randomized Hough: Iteration %d, accumulated maximum %d\n",itercnt,maxaccu);
		if (findcnt>=20000)
			dp (1, "Randomized Hough: Aborted because I couldn't find any more point pairs \n");


	}

	make_minmax (&hough);
	handback (img, &hough);

}




/* Inverse Hough line transform. We assume a Hough transformed image
	where the local maxima have been isolated and the rest thresholded out.
	For each nonzero point, we add a stright line in the destination image
*/


void hough_inverse_line (image_type *img)
{
image_type ihough; float *p;
float deltatheta, theta, rho, buf;
int x1,y1, x,y,z, tx,ty, xm, ym, slcidx, max;
float min, r;


	/* Allocate inverse transform image. Inverse Hough transform is always FLOAT */

	max = IROUND(img->xmax/SQ2);			/* Maximum value for x and y */
	deltatheta = 180.0/img->ymax;		/* Angle increment between lines */
	slcidx = 0;							/* Index into Hough data slice */

	if (allocate_image (&ihough, B_FLOAT, max, max, img->zmax)==-1)
		fatal_memerr ();
	p = ihough.data;
	xm=max/2; ym=max/2;

	for (z=0; z<img->zmax; z++)			/* Run over all slices */
	{
		for (ty=0; ty<img->ymax; ty++)		/* Theta y index */
			for (tx=0; tx<img->xmax; tx++)	/* Rho index */
			{
				if (nonzero(img,tx,ty,z))		/* For all nonzero pixels of this slice */
				{
					/* To trace the line, we first find the point closest to the origin,
						that is xs=rho*cos(theta) and ys=rho*sin(theta) */

					rho = tx - img->xmax/2; theta = ty*deltatheta;
					x = IROUND(rho*COS(theta)); y=IROUND(rho*SIN(theta));
					if ( ((x+xm)>=0) && ((x+xm)<ihough.xmax) && ((y+ym)>=0) && ((y+ym)<ihough.ymax) )
						p[(x+xm) + max*(y+ym) + slcidx] = 1;

					/* Now we begin a forward walk. One of the points +1,+1; 0,+1; +1,0; +1,-1 must 
						match line equation best. Note that these are 4 neighbors in directions
						NE, E, SE, S. */

					while ((x<xm) && (x>-xm) && (y<ym) && (y>-ym))
					{
						x1=x+1; y1=y;
						min = fabs (rho - x1*COS(theta) - y1*SIN(theta));
						if ( (r=fabs (rho - (x+1)*COS(theta) - (y+1)*SIN(theta))) < min)
						{
							x1=x+1; y1=y+1; min=r;
						}
						if ( (r=fabs (rho - (x+1)*COS(theta) - (y-1)*SIN(theta))) < min)
						{
							x1=x+1; y1=y-1; min=r;
						}
						if ( (r=fabs (rho - x*COS(theta) - (y+1)*SIN(theta))) < min)
						{
							x1=x; y1=y+1; min=r;
						}
						x=x1; y=y1; 
						if ( ((x+xm)>=0) && ((x+xm)<ihough.xmax) && ((y+ym)>=0) && ((y+ym)<ihough.ymax) )
							p[(x+xm) + max*(y+ym) + slcidx] = 1;
					}

					/* Start at the same coordinates as above, but walk in the opposite direction.
						This covers the other 4 neighbors in SW, W, NW, N. */

					x = IROUND(rho*COS(theta)); y=IROUND(rho*SIN(theta));
					while ((x<xm) && (x>-xm) && (y<ym) && (y>-ym))
					{
						x1=x-1; y1=y;
						min = fabs (rho - x1*COS(theta) - y1*SIN(theta));
						if ( (r=fabs (rho - (x-1)*COS(theta) - (y+1)*SIN(theta))) < min)
						{
							x1=x-1; y1=y+1; min=r;
						}
						if ( (r=fabs (rho - (x-1)*COS(theta) - (y-1)*SIN(theta))) < min)
						{
							x1=x-1; y1=y-1; min=r;
						}
						if ( (r=fabs (rho - x*COS(theta) - (y-1)*SIN(theta))) < min)
						{
							x1=x; y1=y-1; min=r;
						}
						x=x1; y=y1;
						if ( ((x+xm)>=0) && ((x+xm)<ihough.xmax) && ((y+ym)>=0) && ((y+ym)<ihough.ymax) )
							p[(x+xm) + max*(y+ym) + slcidx] = 1;
					}


				}
			}
		slcidx += SQR(max);			/* next invHough data slice */
	}

	make_minmax (&ihough);
	handback (img, &ihough);

}



/* Forward circle transform. Since circle space is 3D (r,xc,yc),
	this algorithm works only on 2D images.
	Each pixel of the source image is projected into a cone in Hough space.
	This function assumes a pre-thresholded edge image.
	In this function, we restrict the radii to everything smaller than rmax.
*/


void hough_circle (image_type *img, int rmax, int slice)
{
image_type hough; float *p;
float buf;
int tx,ty;
int y,x,z,x1,y1,xb,yb,xc,yc,r,s,e;

	/* Allocate target image. Hough transform is always FLOAT */

	tx = img->xmax + 2*rmax; 
	ty = img->ymax + 2*rmax;

	if (allocate_image (&hough, B_FLOAT, tx, ty, rmax)==-1)
		fatal_memerr ();
	p = hough.data;


	/* Let all pixels of the source image vote */

	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
		{
			if (nonzero(img,x,y,slice))			/* For all nonzero pixels of this slice */
			{
				/* A vote consists of accumulating votes on a cone */

				buf=readbuf_flt(img,x,y,slice); 	/* TODO: Check what this is for */
				p[x+rmax + tx*(y+rmax)] += 1;		/* Circle at r=0 is a point. Much simpler */

				for (r=1; r<rmax; r++)		/* cone height determined by rmax */
				{
					/* Bresenham-circle algorithm ! */
					/* Bresenham-circle does only 1/8 circle, so we need to mirror x8 */

					s=0;			/* step number */
					xb=0; yb=r; 		/* Need to translate to xc,yc later */
					e=r-1;
					z = tx*ty*r;		/* Z offset into Hough image */

					do
					{
						/* Fill all 8 points in the matrix. We move through the 8 segments in CW orientation.
							(xb,yb) is the current Bresenham point
							(x1,y1) is the mirrored/rotated point for the 8 segments
							(xc,yc) is the point shifted to (x,y) as circle center
							Clamp everything that falls outside of the matrix
						*/

						xc = xb+x+rmax; yc = yb+y+rmax;
						if ((xc>=0) && (yc>=0) && (xc<hough.xmax) && (yc<hough.ymax))
							p[ xc + tx*yc + z ] += 1;		/* 1. From y-axis CW */

						x1 = yb; y1 = -xb;
						xc = x1+x+rmax; yc = y1+y+rmax;
						if ((xc>=0) && (yc>=0) && (xc<hough.xmax) && (yc<hough.ymax))
							p[ xc + tx*yc + z ] += 1;		/* 2. From x-axis CCW */

						x1 = yb; y1 = xb;
						xc = x1+x+rmax; yc = y1+y+rmax;
						if ((xc>=0) && (yc>=0) && (xc<hough.xmax) && (yc<hough.ymax))
							p[ xc + tx*yc + z ] += 1;		/* 3. From x-axis CW */

						x1 = xb; y1 = -yb;
						xc = x1+x+rmax; yc = y1+y+rmax;
						if ((xc>=0) && (yc>=0) && (xc<hough.xmax) && (yc<hough.ymax))
							p[ xc + tx*yc + z ] += 1;		/* 4. From neg. y-axis CCW */

						x1 = -xb; y1 = -yb;
						xc = x1+x+rmax; yc = y1+y+rmax;
						if ((xc>=0) && (yc>=0) && (xc<hough.xmax) && (yc<hough.ymax))
							p[ xc + tx*yc + z ] += 1;		/* 5. From neg. y-axis CW */

						x1 = -yb; y1 = xb;
						xc = x1+x+rmax; yc = y1+y+rmax;
						if ((xc>=0) && (yc>=0) && (xc<hough.xmax) && (yc<hough.ymax))
							p[ xc + tx*yc + z ] += 1;		/* 6. From neg. x-axis CCW */

						x1 = -yb; y1 = -xb;
						xc = x1+x+rmax; yc = y1+y+rmax;
						if ((xc>=0) && (yc>=0) && (xc<hough.xmax) && (yc<hough.ymax))
							p[ xc + tx*yc + z ] += 1;		/* 7. From neg. x-axis CW */

						x1 = -xb; y1 = yb;
						xc = x1+x+rmax; yc = y1+y+rmax;
						if ((xc>=0) && (yc>=0) && (xc<hough.xmax) && (yc<hough.ymax))
							p[ xc + tx*yc + z ] += 1;		/* 8. From y-axis CCW */

						/* Bresenham step */

						if (e<0) { yb--; e +=yb+yb; }
						e -= xb+xb+1;
						xb++;
						s++;
					}
					while (xb<=yb);		/* End Bresenham */
				}						/* End cone height (for r) */
			}							/* End for all nonzero pixels */
		}								/* End x/y loop */


	make_minmax (&hough);
	handback (img, &hough);

}




/* Inverse Hough circle transform. We assume a Hough transformed image
	where the local maxima have been isolated and the rest thresholded out.
	For each nonzero point, we add a circle in the target image.
	NOTE: If this is not a 3D image generated by the Hough circle 
	forward transform, the results may be unpredictable.
*/


void hough_inverse_circle (image_type *img)
{
image_type ihough; unsigned char *p;
float buf;
int x1,y1, x,y,z, xm, ym, rmax;
int e,xb,yb,xc,yc,s;


	if (img->zmax<3)
	{
		ShowMessage ("This is probably not a Hough circle transform");
		return;
	}

	/* Allocate imverse transform image. Inverse Hough transform is UCHAR (circle/no circle) */

	rmax = img->zmax;							/* Recover our original rmax parameter */
	xm = img->xmax - 2*rmax;
	ym = img->ymax - 2*rmax;					/* Restore original image's x and y */
	if ( (xm<=0) || (ym<=0))
	{
		ShowMessage ("This is probably not a Hough circle transform");
		return;
	}

	if (allocate_image (&ihough, B_CHAR, xm, ym, 1)==-1)
		fatal_memerr ();
	p = ihough.data;


	for (z=0; z<img->zmax; z++)			/* Run over all slices (=radii) */
	{
		for (y=0; y<img->ymax; y++)			/* y center candidates */
			for (x=0; x<img->xmax; x++)		/* x center candidates */
			{
				if (nonzero(img,x,y,z))		/* For all nonzero pixels of this slice */
				{

					/* Bresenham-circle algorithm ! */
					/* Bresenham-circle does only 1/8 circle, so we need to mirror x8 */

					s=0;			/* step number */
					xb=0; yb=z; 		/* Need to translate to xc,yc later */
					e=z-1;

					do
					{
						/* Fill all 8 points in the matrix. We move through the 8 segments in CW orientation.
							(xb,yb) is the current Bresenham point
							(x1,y1) is the mirrored/rotated point for the 8 segments
							(xc,yc) is the point shifted to (x,y) as circle center
							Clamp everything that falls outside of the matrix
						*/

						xc = xb+x-rmax; yc = yb+y-rmax;
						if ((xc>=0) && (yc>=0) && (xc<ihough.xmax) && (yc<ihough.ymax))
							p[ xc + xm*yc ] = 255;		/* 1. From y-axis CW */

						x1 = yb; y1 = -xb;
						xc = x1+x-rmax; yc = y1+y-rmax;
						if ((xc>=0) && (yc>=0) && (xc<ihough.xmax) && (yc<ihough.ymax))
							p[ xc + xm*yc ] = 255;		/* 2. From x-axis CCW */

						x1 = yb; y1 = xb;
						xc = x1+x-rmax; yc = y1+y-rmax;
						if ((xc>=0) && (yc>=0) && (xc<ihough.xmax) && (yc<ihough.ymax))
							p[ xc + xm*yc ] = 255;		/* 3. From x-axis CW */

						x1 = xb; y1 = -yb;
						xc = x1+x-rmax; yc = y1+y-rmax;
						if ((xc>=0) && (yc>=0) && (xc<ihough.xmax) && (yc<ihough.ymax))
							p[ xc + xm*yc ] = 255;		/* 4. From neg. y-axis CCW */

						x1 = -xb; y1 = -yb;
						xc = x1+x-rmax; yc = y1+y-rmax;
						if ((xc>=0) && (yc>=0) && (xc<ihough.xmax) && (yc<ihough.ymax))
							p[ xc + xm*yc ] = 255;		/* 5. From neg. y-axis CW */

						x1 = -yb; y1 = xb;
						xc = x1+x-rmax; yc = y1+y-rmax;
						if ((xc>=0) && (yc>=0) && (xc<ihough.xmax) && (yc<ihough.ymax))
							p[ xc + xm*yc ] = 255;		/* 6. From neg. x-axis CCW */

						x1 = -yb; y1 = -xb;
						xc = x1+x-rmax; yc = y1+y-rmax;
						if ((xc>=0) && (yc>=0) && (xc<ihough.xmax) && (yc<ihough.ymax))
							p[ xc + xm*yc ] = 255;		/* 7. From neg. x-axis CW */

						x1 = -xb; y1 = yb;
						xc = x1+x-rmax; yc = y1+y-rmax;
						if ((xc>=0) && (yc>=0) && (xc<ihough.xmax) && (yc<ihough.ymax))
							p[ xc + xm*yc ] = 255;		/* 8. From y-axis CCW */

						/* Bresenham step */

						if (e<0) { yb--; e +=yb+yb; }
						e -= xb+xb+1;
						xb++;
						s++;
					}
					while (xb<=yb);		/* End Bresenham */

				}						/* End for all nonzero pixels */
			}							/* End x/y loop */
	}									/* End z (=r) loop */

	make_minmax (&ihough);
	handback (img, &ihough);

}






/********************************************************************

	The Hough transform GUI

*********************************************************************/



GtkAdjustment *hough_deltatheta_adj;
GtkAdjustment *hough_thresh_adj;
GtkWidget *hough_thresh;
GtkWidget *hough_label1, *hough_label2;
GtkAdjustment *hough_radius_adj;
GtkWidget *hough_radius;
int hough_function;



void on_hough_cancelbutton_clicked  (GtkButton *button, gpointer user_data)
{

	gtk_widget_destroy (GTK_WIDGET (user_data));
}


void on_hough_okbutton_clicked  (GtkButton *button, gpointer user_data)
{
float thresh;
int radius;

	undo_nexus (&mainimg);

	if (hough_function==0)		/* Line transform */
	{
		hough_line (&mainimg, hough_deltatheta_adj->value);
		make_imlib_buffer_from_main_image (cur_slice);
	}
	else if (hough_function==1)		/* Line transform with edge detector and orientation limit */
	{
		//thresh = hough_thresh_adj->value; 
		thresh = gtk_spin_button_get_value_as_float ((GtkSpinButton*)hough_thresh);
		dp (2, "Hough: Threshold=%f\n",thresh);
		hough_line_edge (&mainimg, hough_deltatheta_adj->value, thresh, cur_slice);
		make_imlib_buffer_from_main_image (cur_slice=0);
	}
	else if (hough_function==2)		/* Randomized line transform */
	{
		//thresh = hough_thresh_adj->value; 
		thresh = gtk_spin_button_get_value_as_float ((GtkSpinButton*)hough_thresh);
		dp (2, "Hough: Threshold=%f\n",thresh);
		randomized_hough_line (&mainimg, hough_deltatheta_adj->value, cur_slice);
		make_imlib_buffer_from_main_image (cur_slice=0);
	}
	else if (hough_function==3)		/* inverse line transform */
	{
		hough_inverse_line (&mainimg); 
		make_imlib_buffer_from_main_image (cur_slice);
	}
	else if (hough_function==4)		/* circle transform */
	{
		radius = gtk_spin_button_get_value_as_float ((GtkSpinButton*)hough_radius);
		hough_circle (&mainimg, radius, cur_slice); 
		cur_slice=0;
		make_imlib_buffer_from_main_image (cur_slice);
	}
	else if (hough_function==5)		/* inverse circle transform */
	{
		hough_inverse_circle (&mainimg); 
		cur_slice=0;
		make_imlib_buffer_from_main_image (cur_slice);
	}

	redraw (0);

	gtk_widget_destroy (GTK_WIDGET (user_data));
}


void hough_set_menu (GtkWidget *widget, gpointer user_data)
{

	hough_function = *(int*) user_data;
	dp (2, "Hough: Set function %d\n", hough_function);

	if (hough_function==1)
	{
		gtk_widget_show (hough_label1);
		gtk_widget_show (hough_thresh);
		gtk_widget_hide (hough_label2);
		gtk_widget_hide (hough_radius);
	}
	else if (hough_function==4)
	{
		gtk_widget_hide (hough_label1);
		gtk_widget_hide (hough_thresh);
		gtk_widget_show (hough_label2);
		gtk_widget_show (hough_radius);
	}
	else
	{
		gtk_widget_hide (hough_label1);
		gtk_widget_hide (hough_thresh);
		gtk_widget_hide (hough_label2);
		gtk_widget_hide (hough_radius);
	}

}



void make_hough_gui ()
{
  GtkWidget* wnd;
  GtkWidget* vbox;
  GtkWidget* table;
  GtkWidget *hsep;
  GtkWidget *label1;
  GtkWidget *okbutton, *cancelbutton;
  GtkAdjustment *hough_thresh_adj;
  GtkWidget *hough_deltatheta;
  GtkWidget *hough_optionmenu1;
  GtkWidget *hough_optionmenu1_menu;
  GtkWidget *glade_menuitem;
  GtkTooltips *tt_l3d, *tt_le2d, *tt_lr2d, *tt_il3d, *tt_ci2d, *tt_ici2d, *tt_anginc, *tt_ethresh, *tt_maxrad;
  int y;


	wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (wnd), "wnd001", wnd);
	gtk_container_set_border_width (GTK_CONTAINER (wnd), 3);
	gtk_window_set_title (GTK_WINDOW (wnd), "Hough Transform");
	gtk_window_set_modal (GTK_WINDOW (wnd), TRUE);
	vbox = vbox_in_window (wnd,"vbox001");
	table = MakeTable (wnd,vbox,"table001",2,6);

	y=0;
	label1 = MakeLabel (wnd,table,"Geometry","label1",0,y);
	hough_optionmenu1 = Make_Optionmenu (wnd, table,"hough_menu",1,y);
	gtk_container_set_border_width (GTK_CONTAINER (hough_optionmenu1), 1);
	hough_optionmenu1_menu = gtk_menu_new ();
	glade_menuitem = Option_menuitem (hough_optionmenu1_menu, GTK_SIGNAL_FUNC(hough_set_menu), "Line (3D)", (gpointer) &menuopt0);
	tt_l3d = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_l3d, glade_menuitem, "Line Hough transform,\n2D slice by slice for an entire stack", NULL);
	glade_menuitem = Option_menuitem (hough_optionmenu1_menu, GTK_SIGNAL_FUNC(hough_set_menu), "Line w/Edge (2D)", (gpointer) &menuopt1);
	tt_le2d = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_le2d, glade_menuitem, "Line Hough transform with edge direction\n2D, current slice only", NULL);
	glade_menuitem = Option_menuitem (hough_optionmenu1_menu, GTK_SIGNAL_FUNC(hough_set_menu), "Randomized Line (2D)", (gpointer) &menuopt2);
	tt_lr2d = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_lr2d, glade_menuitem, "Randomized line Hough transform,\n(uses randomly selected pixel pairs)\n2D, current slice only", NULL);
	glade_menuitem = Option_menuitem (hough_optionmenu1_menu, GTK_SIGNAL_FUNC(hough_set_menu), "Inverse Line (3D)", (gpointer) &menuopt3);
	tt_il3d = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_il3d, glade_menuitem, "Inverse line Hough transform, valid for all\nof the above functions.\n2D slice-by-slice for an entire stack.", NULL);
	glade_menuitem = Option_menuitem (hough_optionmenu1_menu, GTK_SIGNAL_FUNC(hough_set_menu), "Circle (2D)", (gpointer) &menuopt4);
	tt_ci2d = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_ci2d, glade_menuitem, "Circle Hough transform (creates 3D space)\n2D, current slice only", NULL);
	glade_menuitem = Option_menuitem (hough_optionmenu1_menu, GTK_SIGNAL_FUNC(hough_set_menu), "Inverse Circle (2D)", (gpointer) &menuopt5);
	tt_ici2d = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_ici2d, glade_menuitem, "Inverse circle transform.\n(requires 3D x-y-r space)\nCreates only a single slice (2D reconstruction).", NULL);

	gtk_option_menu_set_menu (GTK_OPTION_MENU (hough_optionmenu1), hough_optionmenu1_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (hough_optionmenu1), 0);
	hough_function=0;

	y++;
	label1 = MakeLabel (wnd,table,"Angle increment","label2",0,y);
	hough_deltatheta_adj = GTK_ADJUSTMENT (gtk_adjustment_new (1, 1, 180, 1, 10, 0));
	hough_deltatheta = MakeHScale (hough_deltatheta_adj,wnd,table,"h_deltatheta",1,y);
	gtk_scale_set_digits (GTK_SCALE (hough_deltatheta), 0);
	tt_anginc = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_anginc, hough_deltatheta, "Angle bins in Hough transform,\nbalances accuracy with robust maximum detection", NULL);

	y++;
	hough_label1 = MakeLabel (wnd,table,"Edge threshold","label3",0,y);
	hough_thresh_adj = GTK_ADJUSTMENT (gtk_adjustment_new (0, -65000, 65000, 1, 10, 0));	/* Precompute? */
	hough_thresh = MakeSpinbutton (GTK_ADJUSTMENT(hough_thresh_adj),wnd,table,"h_thresh",1,y);
	gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (hough_thresh), TRUE);
	tt_ethresh = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_ethresh, hough_thresh, "Threshold at which a pixel is recognized as true edge\n(improves noise suppression)", NULL);
	gtk_widget_hide (hough_label1);
	gtk_widget_hide (hough_thresh);

	y++;
	hough_label2 = MakeLabel (wnd,table,"Maximum radius","label4",0,y);
	hough_radius_adj = GTK_ADJUSTMENT (gtk_adjustment_new (25, 3, mainimg.xmax, 1, 10, 0));
	hough_radius = MakeSpinbutton (GTK_ADJUSTMENT(hough_radius_adj),wnd,table,"h_radius",1,y);
	gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (hough_radius), TRUE);
	tt_maxrad = gtk_tooltips_new ();
	gtk_tooltips_set_tip (tt_maxrad, hough_radius, "Largest circle radius to be considered.\n(limits transform stack size)", NULL);
	gtk_widget_hide (hough_label2);
	gtk_widget_hide (hough_radius);

	y++;
	hsep = MakeHsep (wnd, table, "hhsep1", 0, y);
	hsep = MakeHsep (wnd, table, "hhsep2", 1, y);

	y++;
	okbutton = Button_in_table (wnd,table,"Transform","hough_okbutton",0,y,NULL);
	cancelbutton = Button_in_table (wnd,table,"Cancel","hough_cancelbutton",1,y,NULL);

	gtk_signal_connect (GTK_OBJECT (wnd), "destroy",
                      GTK_SIGNAL_FUNC (on_hough_cancelbutton_clicked), wnd);
	gtk_signal_connect (GTK_OBJECT (cancelbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_hough_cancelbutton_clicked), wnd);
	gtk_signal_connect (GTK_OBJECT (okbutton), "clicked",
                      GTK_SIGNAL_FUNC (on_hough_okbutton_clicked), wnd);

	gtk_widget_show (wnd);

}





void hough_main ()
{

	make_hough_gui ();

}



